/* -*- indent-tabs-mode: t; tab-width: 8; c-basic-offset: 8; -*- */

#include <stdio.h>
#include <assert.h>
#include <unistd.h>
#include <string.h>
#include "options.h"
#include "transport.h"
#include "libburn.h"
#include "drive.h"
#include "sector.h"
#include "crc.h"
#include "debug.h"
#include "lec.h"
#include "toc.h"
#include "write.h"

static unsigned char isrc[] = "0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ";

#define sector_common(X) d->alba++; d->rlba X;\
sector_cook(o, data);

void subcode_toc(struct burn_drive *d, int mode, unsigned char *data)
{
	unsigned char *q;
	int track;
	int crc;
	int min, sec, frame;

	track = d->toc_temp / 3;
	memset(data, 0, 96);
	q = data + 12;

	burn_lba_to_msf(d->rlba, &min, &sec, &frame);
/*XXX track numbers are BCD
a0 - 1st track ctrl
a1 - last track ctrl
a2 - lout ctrl
*/
	q[0] = (d->toc_entry[track].control << 4) + 1;
	q[1] = 0;
	if (d->toc_entry[track].point < 100)
		q[2] = dec_to_bcd(d->toc_entry[track].point);
	else
		q[2] = d->toc_entry[track].point;
	q[3] = dec_to_bcd(min);
	q[4] = dec_to_bcd(sec);
	q[5] = dec_to_bcd(frame);
	q[6] = 0;
	q[7] = dec_to_bcd(d->toc_entry[track].pmin);
	q[8] = dec_to_bcd(d->toc_entry[track].psec);
	q[9] = dec_to_bcd(d->toc_entry[track].pframe);
	crc = crc_ccitt(q, 10);
	q[10] = crc >> 8;
	q[11] = crc & 0xFF;
	d->toc_temp++;
	d->toc_temp %= (d->toc_entries * 3);
}

void sector_toc(struct burn_write_opts *o, int mode)
{
	struct burn_drive *d = o->drive;
	unsigned char data[2352+96];

	memset(data, 0, 2352);
	subcode_toc(d, mode, data + 2352);
	sector_headers(o, data, mode, 1);

	sector_common(++)
}

void subcode_pregap(struct burn_write_opts *o,
                    unsigned char tno, unsigned char control,
                    unsigned char *data)
{
	struct burn_drive *d = o->drive;
	unsigned char *q;
	int rmin, amin, rsec, asec, rframe, aframe, crc;

	memset(data, 0, 96);
	memset(data, 0xFF, 12);	/* p subchannel */
	q = data + 12;

	burn_lba_to_msf(d->alba, &amin, &asec, &aframe);
	burn_lba_to_msf(d->rlba, &rmin, &rsec, &rframe);
/*	burn_print(12, "%d - (%d, %d, %d)", alba, amin, asec, aframe);
	burn_print(12, ": (%d, %d, %d)\n", rmin, rsec, rframe);
*/
/* XXX sometimes other Q modes can be present... */
	q[0] = (control << 4) + 1;
	q[1] = tno;
	q[2] = 0;
	q[3] = dec_to_bcd(rmin);
	q[4] = dec_to_bcd(rsec);
	q[5] = dec_to_bcd(rframe);
	q[6] = 0;
	q[7] = dec_to_bcd(amin);
	q[8] = dec_to_bcd(asec);
	q[9] = dec_to_bcd(aframe);
	crc = crc_ccitt(q, 10);
	q[10] = crc >> 8;
	q[11] = crc & 0xFF;
}

void sector_pregap(struct burn_write_opts *o,
                   unsigned char tno, unsigned char control,
                   int mode)
{
	struct burn_drive *d = o->drive;
	unsigned char data[2352+96];

	memset(data, 0, 2352);
	subcode_pregap(o, tno, control, data + 2352);
	sector_headers(o, data, mode, 0);
	sector_common(--)
}

void subcode_postgap(struct burn_write_opts *o,
                     unsigned char tno, unsigned char control,
                     unsigned char *data)
{
	struct burn_drive *d = o->drive;
	unsigned char *q;

	int crc;
	int rmin, amin, rsec, asec, rframe, aframe;

	memset(data, 0, 96);
	memset(data, 0xFF, 12);	/* p subchannel */
	q = data + 12;

	burn_lba_to_msf(d->alba, &amin, &asec, &aframe);
	burn_lba_to_msf(d->rlba, &rmin, &rsec, &rframe);
/*	burn_print(12, "%d - (%d, %d, %d)", alba, amin, asec, aframe);
	burn_print(12, ": (%d, %d, %d)\n", rmin, rsec, rframe);
*/
/* XXX sometimes other Q modes can be present... */
	q[0] = (control << 4) + 1;
	q[1] = tno;
#warning use current index
	q[2] = 0;
	q[3] = dec_to_bcd(rmin);
	q[4] = dec_to_bcd(rsec);
	q[5] = dec_to_bcd(rframe);
	q[6] = 0;
	q[7] = dec_to_bcd(amin);
	q[8] = dec_to_bcd(asec);
	q[9] = dec_to_bcd(aframe);
	crc = crc_ccitt(q, 10);
	q[10] = crc >> 8;
	q[11] = crc & 0xFF;

}

void sector_postgap(struct burn_write_opts *o,
                    unsigned char tno, unsigned char control,
                    int mode)
{
	struct burn_drive *d = o->drive;
	unsigned char data[2352+96];

	memset(data, 0, 2352);
	memset(data + 2352, 0xFF, 12);	/* p subchannel */

	subcode_postgap(o, tno, control, data + 2352);
	sector_headers(o, data, mode, 0);
	sector_common(++)
}

void subcode_lout(struct burn_write_opts *o, unsigned char control,
                  unsigned char *data)
{
	struct burn_drive *d = o->drive;
	unsigned char *q;
	int crc;
	int rmin, min, rsec, sec, rframe, frame;

	memset(data, 0, 96);
	q = data + 12;

	burn_lba_to_msf(d->alba, &min, &sec, &frame);
	burn_lba_to_msf(d->rlba, &rmin, &rsec, &rframe);

	if (((rmin == 0) && (rsec == 0) && (rframe == 0)) ||
	    ((rsec >= 2) && !((rframe / 19) % 2)))
		memset(data, 0xFF, 12);
	q[0] = (control << 4) + 1;
	q[1] = 0xAA;
	q[2] = 0x01;
	q[3] = dec_to_bcd(rmin);
	q[4] = dec_to_bcd(rsec);
	q[5] = dec_to_bcd(rframe);
	q[6] = 0;
	q[7] = dec_to_bcd(min);
	q[8] = dec_to_bcd(sec);
	q[9] = dec_to_bcd(frame);
	crc = crc_ccitt(q, 10);
	q[10] = crc >> 8;
	q[11] = crc & 0xFF;
}

void sector_lout(struct burn_write_opts *o, unsigned char control,
                 int mode)
{
	struct burn_drive *d = o->drive;
	unsigned char data[2352+96];

	memset(data, 0, 2352);
	subcode_lout(o, control, data + 2352);
	sector_headers(o, data, mode, 0);
	sector_common(++)
}

static char char_to_isrc(char c)
{
	if (c >= '0' && c <= '9')
		return c - '0';
	if (c >= 'A' && c <= 'Z')
		return 0x11 + (c - 'A');
	if (c >= 'a' && c <= 'z')
		return 0x11 + (c - 'a');
	assert(0);
	return 0;
}

void sector_subcodes(struct burn_write_opts *o, unsigned char *subcodes, struct burn_track *t,
                     int prelout)
{
	struct burn_drive *d = o->drive;
	unsigned char *p, *q, *qdata;
	int crc;
	int m, s, f, c, qmode; /* 1, 2 or 3 */

	memset(subcodes, 0, 96);

	p = subcodes;
	if ((t->entry->point == 1) && (d->rlba == -150))
		memset(p, 0xFF, 12);

	if (prelout)
		memset(p, 0xFF, 12);
	q = subcodes + 12;

	qmode = 1;
	/* every 1 in 10 we can do something different */
	if (d->rlba % 10 == 0) {
		/* each of these can occur 1 in 100 */
		if ((d->rlba / 10) % 10 == 0) {
			if (o->has_mediacatalog)
				qmode = 2;
		} else if ((d->rlba / 10) % 10 == 1) {
			if (t->has_isrc)
				qmode = 3;
		}
	}

	assert(qmode == 1 || qmode == 2 || qmode == 3);

	qdata = q + 1;
	switch (qmode) {
	case 1:
		qdata[0] = dec_to_bcd(t->entry->point);	/* track number */
		qdata[1] = dec_to_bcd(1);	/* index XXX read this shit
						   from the track array */
		burn_lba_to_msf(d->rlba, &m, &s, &f);
		qdata[2] = dec_to_bcd(m);	/* rel min */
		qdata[3] = dec_to_bcd(s);	/* rel sec */
		qdata[4] = dec_to_bcd(f);	/* rel frame */
		qdata[5] = 0;	/* zero */
		burn_lba_to_msf(d->alba, &m, &s, &f);
		qdata[6] = dec_to_bcd(m);	/* abs min */
		qdata[7] = dec_to_bcd(s);	/* abs sec */
		qdata[8] = dec_to_bcd(f);	/* abs frame */
		break;
	case 2:
		/* media catalog number */
		qdata[0] =
			(o->mediacatalog[0] << 4) +
			o->mediacatalog[1];
		qdata[1] =
			(o->mediacatalog[2] << 4) +
			o->mediacatalog[3];
		qdata[2] =
			(o->mediacatalog[4] << 4) +
			o->mediacatalog[5];
		qdata[3] =
			(o->mediacatalog[6] << 4) +
			o->mediacatalog[7];
		qdata[4] =
			(o->mediacatalog[8] << 4) +
			o->mediacatalog[9];
		qdata[5] =
			(o->mediacatalog[10] << 4) +
			o->mediacatalog[11];
		qdata[6] = o->mediacatalog[12] << 4;

		qdata[7] = 0;  
		burn_lba_to_msf(d->alba, &m, &s, &f);
		qdata[8] = dec_to_bcd(f);	 /* abs frame */
		break;
	case 3:
		c = char_to_isrc(t->country[0]);
		/* top 6 bits of [0] is the first country code */
		qdata[0] = c << 2;
		c = char_to_isrc(t->country[1]);
		/* bottom 2 bits of [0] is part of the second country code */
		qdata[0] += (c >> 4);
		/* top 4 bits if [1] is the rest of the second country code */
		qdata[1] = c << 4;

		c = char_to_isrc(t->owner[0]);
		/* bottom 4 bits of [1] is part of the first owner code */
		qdata[1] += (c >> 2);
		/* top 2 bits of [2] is the rest of the first owner code */
		qdata[2] = c << 6;
		c = char_to_isrc(t->owner[1]);
		/* bottom 6 bits of [2] is the entire second owner code */
		qdata[2] += c;
		c = char_to_isrc(t->owner[2]);
		/* top 6 bits of [3] are the third owner code */
		qdata[3] = c << 2;

		/* [4] is the year in 2 BCD numbers */
		qdata[4] = dec_to_bcd(t->year % 100);
		/* [5] is the first 2 digits in the serial */
		qdata[5] = dec_to_bcd(t->serial % 100);
		/* [6] is the next 2 digits in the serial */
		qdata[6] = dec_to_bcd((t->serial / 100) % 100);
		/* the top 4 bits of [7] is the last serial digit, the rest is 
		   zeros */
		qdata[7] = dec_to_bcd((t->serial / 10000) % 10) << 4;
		burn_lba_to_msf(d->alba, &m, &s, &f);
		qdata[8] = dec_to_bcd(f);	/* abs frame */
		break;
	}
	q[0] = (t->entry->control << 4) + qmode;

	crc = crc_ccitt(q, 10);
	q[10] = crc >> 8;
	q[11] = crc & 0xff;
}

void sector_data(struct burn_write_opts *o, struct burn_track *t, int prelout)
{
	struct burn_drive *d = o->drive;
	unsigned char data[2352+96];
	int offset = -1;

	if (t->mode & BURN_SECTORS)
		offset = 0;
	else if (t->mode & BURN_BYTES) {
		if (t->mode & BURN_AUDIO)
			offset = 0;
		else if (t->mode & BURN_MODE1)
			offset = 16;
	}
	assert(offset != -1);
	t->source->read(t->source, data + offset,
	     burn_sector_length(t->mode));

	if (t->mode & BURN_BYTES) {
		if (t->mode & BURN_MODE1)
			sector_headers(o, data, t->mode, 0);
	}
/* XXX - use the sub modifier stuff here */
	if (t->source->read_sub)
		t->source->read_sub(t->source, data + 2352, 96);
	else
		sector_subcodes(o, data + 2352, t, prelout);

	sector_common(++)
}

/*
	convert and add a sector to a page
	returns 0 when the page can hold no more
	input sectors are always 2352/96

	note: because we don't know how big the next sector is, we
	      return 0 if we can't fit a 2352/96 sector
*/
int sector_cook(struct burn_write_opts *opts, unsigned char *in)
{
	struct burn_drive *d = opts->drive;
	struct buffer *out = d->buffer;
	unsigned char *i, *o;
	int seclen;

	switch (opts->block_type) {
	case BURN_BLOCK_RAW0:
		seclen = 2352;
		memcpy(out->data + out->bytes, in, seclen);
		break;

	case BURN_BLOCK_RAW96R:
		seclen = 2352 + 96;
		memcpy(out->data + out->bytes, in, seclen);
		break;

	case BURN_BLOCK_RAW16:
		memcpy(out->data + out->bytes, in, 2352);
		memcpy(out->data + out->bytes + 2352, in + 2352 + 12, 12);
		o = out->data + out->bytes + 2352 + 12;
		o[0] = 0;
		o[1] = 0;
		o[2] = 0;
/*XXX find a better way to deal with partially damaged P channels*/
		i = in + 2352;
		if (i[2] != 0)
			o[3] = 0x80;
		else
			o[3] = 0;
/*XXX do I really flip these?*/
		o = out->data + out->bytes + 2352 + 10;
		o[0] = ~o[0];
		o[1] = ~o[1];

		seclen = 2352 + 16;
		break;
	default:
		assert(0);
	}

	out->bytes += seclen;
	out->sectors++;

	if (out->bytes + (2352+96) >= BUFFER_SIZE) {
		d->write(d, d->nwa, out);
		d->nwa += out->sectors;
		out->bytes = 0;
		out->sectors = 0;
		return 0;
	}
	return 1;
}

int burn_msf_to_lba(int m, int s, int f)
{
	if (m < 90)
		return (m * 60 + s) * 75 + f - 150;
	else
		return (m * 60 + s) * 75 + f - 450150;
}

void burn_lba_to_msf(int lba, int *m, int *s, int *f)
{
	if (lba >= -150) {
		*m = (lba + 150) / (60 * 75);
		*s = (lba + 150 - *m * 60 * 75) / 75;
		*f = lba + 150 - *m * 60 * 75 - *s * 75;
	} else {
		*m = (lba + 450150) / (60 * 75);
		*s = (lba + 450150 - *m * 60 * 75) / 75;
		*f = lba + 450150 - *m * 60 * 75 - *s * 75;
	}
}

int dec_to_bcd(int d)
{
	int top, bottom;

	top = d / 10;
	bottom = d - (top * 10);
	return (top << 4) + bottom;
}

void sector_headers(struct burn_write_opts *o, unsigned char *out,
                    int mode, int leadin)
{
	struct burn_drive *d = o->drive;
	unsigned int crc;
	int min, sec, frame;
	int modebyte = -1;

	if (mode & BURN_AUDIO)	/* no headers for "audio" */
		return;
	if (mode & BURN_MODE1)
		modebyte = 1;

	assert(modebyte == 1);

	out[0] = 0;
	memset(out + 1, 0xFF, 10);	/* sync */
	out[11] = 0;

	if (leadin) {
		burn_lba_to_msf(d->rlba, &min, &sec, &frame);
		out[12] = dec_to_bcd(min) + 0xA0;
		out[13] = dec_to_bcd(sec);
		out[14] = dec_to_bcd(frame);
		out[15] = modebyte;
	} else {
		burn_lba_to_msf(d->alba, &min, &sec, &frame);
		out[12] = dec_to_bcd(min);
		out[13] = dec_to_bcd(sec);
		out[14] = dec_to_bcd(frame);
		out[15] = modebyte;
	}
	if (mode & BURN_MODE1) {
		crc = crc_32(out, 2064);
		out[2064] = crc & 0xFF;
		crc >>= 8;
		out[2065] = crc & 0xFF;
		crc >>= 8;
		out[2066] = crc & 0xFF;
		crc >>= 8;
		out[2067] = crc & 0xFF;
	}
	if (mode & BURN_MODE1) {
		memset(out + 2068, 0, 8);
		parity_p(out);
		parity_q(out);
	}
	scramble(out);
}

#if 0
void process_q(struct burn_drive *d, unsigned char *q)
{
	unsigned char i[5];
	int mode;

	mode = q[0] & 0xF;
/*      burn_print(12, "mode: %d : ", mode);*/
	switch (mode) {
	case 1:
/*              burn_print(12, "tno = %d : ", q[1]);
                burn_print(12, "index = %d\n", q[2]);
*/
		/* q[1] is the track number (starting at 1) q[2] is the index
		   number (starting at 0) */
#warning this is totally bogus
if (q[1]-1 > 99) break;
		if (q[2] > d->toc->track[q[1] - 1].indices) {
			burn_print(12, "new index at %d\n", d->alba);
			d->toc->track[q[1] - 1].index[q[2]] = d->alba;
			d->toc->track[q[1] - 1].indices++;
		}
		break;
	case 2:
		/* XXX dont ignore these */
		break;
	case 3:
/*              burn_print(12, "ISRC data in mode 3 q\n");*/
		i[0] = isrc[(q[1] << 2) >> 2];
/*              burn_print(12, "0x%x 0x%x 0x%x 0x%x 0x%x\n", q[1], q[2], q[3], q[4], q[5]);
                burn_print(12, "ISRC - %c%c%c%c%c\n", i[0], i[1], i[2], i[3], i[4]);
*/
		break;
	default:
		assert(0);
	}
}
#endif

/* this needs more info.  subs in the data? control/adr? */
#warning sector_identify needs to be written
int sector_identify(unsigned char *data)
{
	scramble(data);
/*
check mode byte for 1 or 2
test parity to see if it's a valid sector
if invalid, return BURN_MODE_AUDIO;
else return mode byte  (what about mode 2 formless?  heh)
*/
	return BURN_MODE1;
}
